@@ -1,8 +1,8 @@
-/*  $Id$
+/*  $Id: CosmoSmashEngine.cpp,v 1.20 2002/12/27 02:52:47 sarrazip Exp $
     CosmoSmashEngine.cpp - A space rock shooting video game engine.
 
     cosmosmash - A space rock shooting video game.
-    Copyright (C) 2000, 2001 Pierre Sarrazin <sarrazip@iname.com>
+    Copyright (C) 2000, 2001, 2002 Pierre Sarrazin <http://sarrazip.com/>
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -20,10 +20,6 @@
     02111-1307, USA.
 */
 
-#ifndef PACKAGE
-    #define PACKAGE "cosmosmash"
-#endif
-
 #include "CosmoSmashEngine.h"
 
 #include <gengameng/PixmapLoadError.h>
@@ -42,6 +38,7 @@
 #include <iomanip>
 
 #include "base.xpm"
+#include "boosted-base.xpm"
 #include "basebullet.xpm"
 #include "bigrock0.xpm"
 #include "bigrock1.xpm"
@@ -73,12 +70,13 @@
 #include "saucerbullet_empty.xpm"
 #include "saucerbullet_mid.xpm"
 #include "saucerbullet_full.xpm"
+#include "question0.xpm"
+#include "question1.xpm"
 
 
 #define __(s) (s)  /* Gettext-like macro */
 
 #ifdef _MSC_VER
-    #define VERSION ""
     #define snprintf _snprintf
 #endif
 
@@ -97,9 +95,11 @@ static const Couple theDrawingPixmapSize(675, 520);
 
 static const int
 PLAYER_SPEED = 15,
+PLAYER_BULLET_SPEED = 36,
 PULSAR_SPEED = 20,
 SAUCER_SPEED = 8,
-SAUCER_BULLET_SPEED = 20;
+SAUCER_BULLET_SPEED = 20,
+QUESTION_SPEED = 16;
 
 static const int
 SAUCER_Y_POS = 75,
@@ -117,7 +117,8 @@ SCORE_BIG_SPINNER = 40,
 SCORE_SMALL_SPINNER = 80,
 SCORE_PULSAR = 50,
 SCORE_SAUCER_BULLET = 50,
-SCORE_SAUCER = 100;
+SCORE_SAUCER = 100,
+SCORE_QUESTION = 50;
 
 
 // Defined as constants for efficiency:
@@ -228,7 +229,7 @@ decrementAmplitude(const Couple &c)
 }
 
 
-static void
+inline void
 removeNullElementsFromSpriteList(SpriteList &slist)
 {
     SpriteList::iterator it =
@@ -240,37 +241,28 @@ removeNullElementsFromSpriteList(SpriteList &slist)
 	    @sarrazip 20010501
 	*/
     slist.erase(it, slist.end());
-
-    #if 0  /* Obsolete @sarrazip 20010501 */
-    for (SpriteList::iterator it = slist.end(); it-- != slist.begin(); )
-	if (*it == NULL)
-	    slist.erase(it);
-    #endif
 }
 
 
 ///////////////////////////////////////////////////////////////////////////////
 
 
-static string
-getWMCaption()
-{
-    return string("Cosmosmash ") + VERSION;
-}
-
-
 CosmoSmashEngine::CosmoSmashEngine(
-			long initialScore, bool _mirrorHyperspace)
+		    long initialScore,
+		    bool _mirrorHyperspace, bool useSound, bool fullScreen)
 							throw(int, string)
-  : GameEngine(theDrawingPixmapSize, getWMCaption(), false),
+  : GameEngine(theDrawingPixmapSize, "Cosmosmash", fullScreen),
   		// may throw string exception
 
     paused(false),
     tickCount(0),
 
     mirrorHyperspace(_mirrorHyperspace),
+    useGameExtensions(false),
+    playerBoostTicks(0),
+    playerBoostType(NO_BOOST),
 
-    playerPA(1),
+    playerPA(2),
     playerSprite(NULL),
 
     baseBulletPA(1),
@@ -296,6 +288,13 @@ CosmoSmashEngine::CosmoSmashEngine(
     explosionPA(2),
     explosionSprites(),
 
+    questionPA(2),
+    questionSprites(),
+
+    groundPos(0),
+    groundHeight(0),
+    mountainTopPos(0),
+
     leftArrowPressed(false),
     rightArrowPressed(false),
     escapePressed(false),
@@ -303,6 +302,8 @@ CosmoSmashEngine::CosmoSmashEngine(
     spacePressedBefore(false),
     downArrowPressed(false),
     downArrowPressedBefore(false),
+    letterEPressed(false),
+    letterEPressedBefore(false),
     letterPPressed(false),
     letterPPressedBefore(false),
     ctrlPressed(false),
@@ -317,8 +318,11 @@ CosmoSmashEngine::CosmoSmashEngine(
 
     timeBeforeNextPulsar(0),
     timeBeforeNextSaucer(0),
+    timeBeforeNextQuestion(0),
 
-    starPositions()
+    starPositions(),
+
+    theSoundMixer(NULL)
 {
     assert(initScore >= 0);
     assert(theSDLScreen != NULL);
@@ -330,30 +334,15 @@ CosmoSmashEngine::CosmoSmashEngine(
 
     try
     {
-	try
-	{
-	    initializeSprites();
-	}
-	catch (PixmapLoadError &e)
-	{
-	    string msg = __("Could not load pixmap ") + e.getFilename();
-	    throw msg;
-	}
-
-	initializeMisc();  // can throw a 'string'
+	initializeSprites();
     }
-    catch (string &msg)
+    catch (PixmapLoadError &e)
     {
-	displayErrorMessage(msg);
-	throw -1;
+	string msg = __("Could not load pixmap ") + e.getFilename();
+	throw msg;
     }
-}
 
-
-void
-CosmoSmashEngine::displayErrorMessage(const string &msg) throw()
-{
-    writeString(msg, Couple(10, 20));
+    initializeMisc(useSound);  // may throw a 'string'
 }
 
 
@@ -371,6 +360,7 @@ CosmoSmashEngine::initializeSprites() throw(PixmapLoadError)
 
     // Load some pixmaps:
     loadPixmap(base_xpm, playerPA, 0);
+    loadPixmap(boosted_base_xpm, playerPA, 1);
 
     // Create and position the player sprite:
     playerSprite = new Sprite(playerPA,
@@ -427,11 +417,33 @@ CosmoSmashEngine::initializeSprites() throw(PixmapLoadError)
     // Explosion:
     loadPixmap(explosion0_xpm, explosionPA, 0);
     loadPixmap(explosion1_xpm, explosionPA, 1);
+
+    // Question (surprise) sprites:
+    loadPixmap(question0_xpm, questionPA, 0);
+    loadPixmap(question1_xpm, questionPA, 1);
+}
+
+
+static
+string
+getDir(const char *defaultValue, const char *envVarName)
+{
+    string dir;
+    const char *s = getenv(envVarName);
+    if (s != NULL)
+	dir = s;
+    else
+	dir = defaultValue;
+    
+    if (!dir.empty() && dir[dir.length() - 1] != '/')
+	dir += '/';
+
+    return dir;
 }
 
 
 void
-CosmoSmashEngine::initializeMisc() throw(string)
+CosmoSmashEngine::initializeMisc(bool useSound) throw(string)
 /*  
     Throws an error message in a 'string' if an error occurs.
 */
@@ -443,6 +455,8 @@ CosmoSmashEngine::initializeMisc() throw(string)
     spacePressedBefore = false;
     downArrowPressed = false;
     downArrowPressedBefore = false;
+    letterEPressed = false;
+    letterEPressedBefore = false;
     letterPPressed = false;
     letterPPressedBefore = false;
     ctrlPressed = false;
@@ -455,15 +469,23 @@ CosmoSmashEngine::initializeMisc() throw(string)
     numLives = 0;
     updateNumLives = true;
     numLivesAreaPos = Couple(theDrawingPixmapSize.x - 72, scoreAreaPos.y);
+    playerBoostTicks = 0;
+    playerBoostType = NO_BOOST;
 
 
     setTimeBeforeNextPulsar();
     setTimeBeforeNextSaucer();
+    setTimeBeforeNextQuestion();
+
+
     /*  Initialize mountain range positions:
     */
     int y0 = groundPos - 100;
     int y1 = groundPos -  70;
     int y2 = groundPos -  50;
+
+    mountainTopPos = y0;
+
     size_t i;
     for (i = 0; i < NUM_MOUNTAIN_RANGE_POINTS; i++)
     {
@@ -486,6 +508,34 @@ CosmoSmashEngine::initializeMisc() throw(string)
 	int y = rnd(padding, y2 - 10);
 	starPositions.push_back(Couple(x, y));
     }
+
+
+    /*  Sound effects:
+	(ignored for now until good sound effects are found @sarrazip 20020330)
+    */
+    if (useSound)
+    {
+	try
+	{
+	    string pkgsounddir = getDir(PKGSOUNDDIR, "PKGSOUNDDIR");
+	    theSoundMixer = new SoundMixer();  // may throw string
+	    for (size_t i = 0; i < NUM_ROCK_HIT_SOUNDS; i++)
+	    {
+		string filename =
+			    pkgsounddir + "rock-hit-" + char('0' + i) + ".wav";
+		rockHitSounds[i].init(filename);
+	    }
+
+	    playerHitSound.init(pkgsounddir + "player-hit.wav");
+	    pulsarBeepSound.init(pkgsounddir + "pulsar-beep.wav");
+	    saucerShootingSound.init(pkgsounddir + "saucer-shooting.wav");
+	    hyperspaceSound.init(pkgsounddir + "hyperspace.wav");
+	}
+	catch (const SoundMixer::Error &e)
+	{
+	    throw e.what();
+	}
+    }
 }
 
 
@@ -505,6 +555,14 @@ CosmoSmashEngine::setTimeBeforeNextSaucer()
 }
 
 
+void
+CosmoSmashEngine::setTimeBeforeNextQuestion()
+{
+    static const char *q = getenv("Q");
+    timeBeforeNextQuestion = (q != NULL ? 100 : rnd(20 * 10, 20 * 30));
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 
@@ -518,6 +576,7 @@ CosmoSmashEngine::processKey(SDLKey keysym, bool pressed)
         case SDLK_RIGHT:   rightArrowPressed = pressed; break;
         case SDLK_DOWN:    downArrowPressed = pressed; break;
         case SDLK_SPACE:   spacePressed = pressed; break;
+        case SDLK_e:       letterEPressed = pressed; break;
         case SDLK_p:       letterPPressed = pressed; break;
 	case SDLK_LCTRL:
 	case SDLK_RCTRL:   ctrlPressed = pressed; break;
@@ -535,8 +594,6 @@ CosmoSmashEngine::tick()
 
     if (paused)
     {
-	restoreBackground();
-	drawSprites();
 	if (letterPPressed && !letterPPressedBefore)
 	{
 	    paused = false;
@@ -555,16 +612,92 @@ CosmoSmashEngine::tick()
 	animateAutomaticCharacters();
 	restoreBackground();
 	drawSprites();
+
+
+	/*  If the game is in demo mode, the Space bar starts a new game.
+	*/
+	if (numLives == 0)
+	{
+	    if (!spacePressed && spacePressedBefore
+			|| !letterEPressed && letterEPressedBefore)
+	    {
+		displayStartMessage(false);
+
+		useGameExtensions = letterEPressedBefore;
+
+		killAllAutomaticCharacters();
+		centerPlayerSprite();
+		theScore = initScore;
+		thePeakScore = initScore;
+		numLives = NUM_INIT_LIVES;
+
+		addToScore(0);
+		addToNumLives(0);
+	    }
+	    else 
+	    {
+		/*  At the end of a game, wait for the last base to have
+		    finished exploding before displaying the start
+		    message again.
+		*/
+		if (explosionSprites.size() == 0)
+		    displayStartMessage(true);
+	    }
+	}
+
+	if (paused)
+	    displayPauseMessage(true);
     }
 
     spacePressedBefore = spacePressed;
     downArrowPressedBefore = downArrowPressed;
+    letterEPressedBefore = letterEPressed;
     letterPPressedBefore = letterPPressed;
 
     return true;
 }
 
 
+void
+CosmoSmashEngine::fireBullet(Couple pos)
+{
+    Couple bbSize = baseBulletPA.getImageSize();
+    Sprite *s = new Sprite(baseBulletPA,
+		pos, Couple(0, -PLAYER_BULLET_SPEED), Couple(0, 0),
+		Couple(0, 0), bbSize);
+    baseBulletSprites.push_back(s);
+
+    if (theLevel >= 5)  // double bullet in higher levels
+    {
+	int n = (theLevel == 5 ? bbSize.y / 2 : bbSize.y);
+	Couple pos2 = pos;
+	pos2.y -= n;
+	Sprite *s = new Sprite(baseBulletPA,
+		pos2, Couple(0, -PLAYER_BULLET_SPEED), Couple(0, 0),
+		Couple(0, 0), bbSize);
+	baseBulletSprites.push_back(s);
+    }
+
+    if (playerBoostTicks > 0 && playerBoostType == TRIPLE_BULLETS)
+    {
+	int vx = 3;
+	Couple pos2 = pos;
+	pos2.x -= bbSize.x + 1;
+	Sprite *s = new Sprite(baseBulletPA,
+		    pos2, Couple(-vx, -PLAYER_BULLET_SPEED), Couple(0, 0),
+		    Couple(0, 0), bbSize);
+	baseBulletSprites.push_back(s);
+
+	pos2 = pos;
+	pos2.x += bbSize.x + 1;
+	s = new Sprite(baseBulletPA,
+		    pos2, Couple(+vx, -PLAYER_BULLET_SPEED), Couple(0, 0),
+		    Couple(0, 0), bbSize);
+	baseBulletSprites.push_back(s);
+    }
+}
+
+
 bool
 CosmoSmashEngine::animatePlayer()
 /*  
@@ -574,7 +707,6 @@ CosmoSmashEngine::animatePlayer()
     if (letterPPressed && !letterPPressedBefore)
     {
 	paused = true;
-	displayPauseMessage(true);
 	return true;
     }
 
@@ -586,27 +718,8 @@ CosmoSmashEngine::animatePlayer()
     }
 
     if (numLives == 0)
-    {
-	/*  If the game is in demo mode, the Space bar starts a new game.
-	*/
-	if (!spacePressed && spacePressedBefore)
-	{
-	    displayStartMessage(false);
-
-	    killAllMeanies();
-	    centerPlayerSprite();
-	    theScore = initScore;
-	    thePeakScore = initScore;
-	    numLives = NUM_INIT_LIVES;
-
-	    addToScore(0);
-	    addToNumLives(0);
-	}
-	else
-	    displayStartMessage(true);
-
 	return true;
-    }
+
 
     Couple &playerPos = playerSprite->getPos();
     const Couple &playerSize = playerSprite->getSize();
@@ -614,26 +727,20 @@ CosmoSmashEngine::animatePlayer()
 
     /*  Fire a bullet once in a while:
     */
-    if (!ctrlPressed && tickCount % 5 == 0)
+    int ticksBetweenBullets =
+	   (playerBoostTicks > 0 && playerBoostType == CLOSER_BULLETS ? 3 : 6);
+    if (tickCount % ticksBetweenBullets == 0)
     {
-	Couple bbSize = baseBulletPA.getImageSize();
-	Couple pos = playerPos
-		+ Couple((playerSize.x - bbSize.x) / 2, - bbSize.y);
-	Sprite *s = new Sprite(baseBulletPA,
-				    pos, Couple(0, -37), Couple(0, 0),
-				    Couple(0, 0), bbSize);
-	baseBulletSprites.push_back(s);
-
-	if (theLevel >= 5)  // double bullet in higher levels
+	if (!ctrlPressed)
 	{
-	    int n = (theLevel == 5 ? bbSize.y / 2 : bbSize.y);
-	    pos.y -= n;
-	    Sprite *s = new Sprite(baseBulletPA,
-					pos, Couple(0, -37), Couple(0, 0),
-					Couple(0, 0), bbSize);
-	    baseBulletSprites.push_back(s);
+	    Couple bbSize = baseBulletPA.getImageSize();
+	    Couple pos = playerPos + Couple((playerSize.x - bbSize.x) / 2, 0);
+
+	    fireBullet(pos);
 	}
     }
+    if (playerBoostTicks > 0)
+	playerBoostTicks--;
 
 
     Couple &playerSpeed = playerSprite->getSpeed();
@@ -661,14 +768,16 @@ CosmoSmashEngine::animatePlayer()
 
 	Couple newPos(newX, groundPos - playerPA.getImageSize().y);
 	repositionPlayerSprite(newPos);
+
+	playSoundEffect(hyperspaceSound);
+
 	return true;
     }
 
-    const int playerSpeedFactor = PLAYER_SPEED;
     if (leftArrowPressed)
-	playerSpeed.x = -playerSpeedFactor;
+	playerSpeed.x = -PLAYER_SPEED;
     if (rightArrowPressed)
-	playerSpeed.x = +playerSpeedFactor;
+	playerSpeed.x = +PLAYER_SPEED;
 
     if (playerSpeed.isZero())
 	return true;
@@ -679,8 +788,6 @@ CosmoSmashEngine::animatePlayer()
     else if (newPos.x + playerSize.x > theDrawingPixmapSize.x)
 	newPos.x = theDrawingPixmapSize.x - playerSize.x;
 
-    // erase the player's sprite before moving its position:
-    eraseSprite(*playerSprite);
     playerPos = newPos;
 
     return true;
@@ -691,10 +798,7 @@ void
 CosmoSmashEngine::killSpritesInList(SpriteList &sl)
 {
     for (SpriteList::iterator its = sl.begin(); its != sl.end(); its++)
-    {
-	eraseSprite(**its);
 	delete *its;
-    }
 
     sl.clear();
     assert(sl.size() == 0);
@@ -714,19 +818,19 @@ CosmoSmashEngine::centerPlayerSprite()
 void
 CosmoSmashEngine::repositionPlayerSprite(Couple newPos)
 {
-    eraseSprite(*playerSprite);
     playerSprite->setPos(newPos);
 }
 
 
 void
-CosmoSmashEngine::killAllMeanies()
+CosmoSmashEngine::killAllAutomaticCharacters()
 {
     killSpritesInList(rockSprites);
     killSpritesInList(spinnerSprites);
     killSpritesInList(pulsarSprites);
     killSpritesInList(saucerSprites);
     killSpritesInList(saucerBulletSprites);
+    killSpritesInList(questionSprites);
 }
 
 
@@ -738,6 +842,8 @@ CosmoSmashEngine::killPlayerBase()
     if (noKillEnvVar != NULL)
     	return;
 
+    playSoundEffect(playerHitSound);
+
     addToScore(SCORE_LOST_BASE);
 
     Couple pos = playerSprite->getCenterPos();
@@ -769,11 +875,16 @@ CosmoSmashEngine::killPlayerBase()
     centerPlayerSprite();
     addToNumLives(-1);
 
-    killAllMeanies();
+    playerBoostTicks = 0;
+    playerBoostType = NO_BOOST;
+
+    killAllAutomaticCharacters();
 	    /*  We kill falling sprites to avoid situations where e.g.,
 		spinners keep landing while the player is exploding.
 		Such situations could make the player lose many lives in a row.
 	    */
+
+    killSpritesInList(baseBulletSprites);
 }
 
 
@@ -812,7 +923,6 @@ CosmoSmashEngine::animateFallingObjects(SpriteList &sprites)
     {
 	Sprite *s = *it;
 	assert(s != NULL);
-	eraseSprite(*s);
 	s->addSpeedToPos();
 
 	if (playerIsActive() && s->collidesWithSprite(*playerSprite))
@@ -864,7 +974,6 @@ CosmoSmashEngine::animateSaucers()
 	Couple &pos = saucer->getPos();
 	const Couple &size = saucer->getSize();
 
-	eraseSprite(*saucer);
 	pos += saucer->getSpeed();
 
 	// if the saucer is now off the screen:
@@ -892,6 +1001,8 @@ CosmoSmashEngine::animateSaucers()
 				    pos, speed, Couple(0, 0),
 				    Couple(2, 2), size - Couple(2, 2));
 	    saucerBulletSprites.push_back(s);
+
+	    playSoundEffect(saucerShootingSound);
 	}
     }
     removeNullElementsFromSpriteList(saucerSprites);
@@ -987,6 +1098,30 @@ CosmoSmashEngine::animateAutomaticCharacters()
     }
 
 
+    /*  Send a new question mark once in a while.
+    */
+    if (useGameExtensions
+		&& theLevel >= 0
+		&& playerSprite->getTimeToLive() == 0
+		&& --timeBeforeNextQuestion == 0)
+    {
+	setTimeBeforeNextQuestion();
+	const Couple size = questionPA.getImageSize();
+
+	Couple pos, speed;
+	pos = Couple(
+		rnd(size.x * 4, theScreenSizeInPixels.x - size.x * 4),
+		- size.y);
+	speed = Couple(rnd(-2, +2), QUESTION_SPEED + 2 * (theLevel >= 5));
+
+	Sprite *s = new Sprite(questionPA,
+				pos, speed, Couple(0, 0),
+				Couple(2, 2), size - Couple(2, 2));
+	s->currentPixmapIndex = 0;
+	questionSprites.push_back(s);
+    }
+
+
     /*  Let the rocks fall. If they hit the ground or go off the screen
 	on the left or right, kill them. If they hit the player (while it
 	is not currently exploding), then make the player explode.
@@ -999,6 +1134,17 @@ CosmoSmashEngine::animateAutomaticCharacters()
 	killPlayerBase();
 
 
+    /*  Let the question marks fall...
+    */
+    if (useGameExtensions)
+    {
+	pair<bool, size_t> p = animateFallingObjects(questionSprites);
+	// We don't care when the question marks land.
+	if (p.first)  // if collision with player
+	    killPlayerBase();
+    }
+
+
     /*  Let the spinners fall. If they hit the ground, the player dies.
 	If they go off the screen on the left or right, kill them.
 	If they hit the player (while it is not currently exploding),
@@ -1048,7 +1194,7 @@ CosmoSmashEngine::animateAutomaticCharacters()
 	{
 	    Couple diff = playerSprite->getPos() - pulsar->getPos();
 	    // NOTE: use projection instead of trig; see animateSaucers()
-	    double angle = atan2(diff.y, diff.x);
+	    double angle = atan2((double) diff.y, (double) diff.x);
 	    Couple newSpeed(int(PULSAR_SPEED * cos(angle)),
 					int(PULSAR_SPEED * sin(angle)));
 	    if (newSpeed.isZero())
@@ -1070,7 +1216,6 @@ CosmoSmashEngine::animateAutomaticCharacters()
     {
 	Sprite *s = *it;
 	assert(s != NULL);
-	eraseSprite(*s);
 	s->addSpeedToPos();
 	if (s->getPos().y + s->getSize().y <= 0)  // if completely above screen
 	{
@@ -1094,7 +1239,6 @@ CosmoSmashEngine::animateAutomaticCharacters()
 	    					s->currentPixmapIndex);
 	    assert(false);
 	}
-	eraseSprite(*s);
 	if (s->decTimeToLive() == 0)
 	{
 	    delete s;
@@ -1213,9 +1357,9 @@ CosmoSmashEngine::makeSpriteExplode(const Sprite *target,
     assert(target != NULL);
     assert(bullet != NULL);
 
+    playShortExplosionSound();
+
     Couple targetPos = target->getPos();
-    eraseSprite(*bullet);
-    eraseSprite(*target);
 
     Couple expPos = target->getCenterPos();
     createExplosionSprite(expPos, 6);
@@ -1236,7 +1380,7 @@ CosmoSmashEngine::detectCollisionsWithBullet(const Sprite &bullet,
     If a collision is detected between 'bullet' and an element
     of 'slist', this element is made null in 'slist' and the
     corresponding Sprite object is destroyed with operator delete;
-    also, an explosion is created at the point of collsion and the
+    also, an explosion is created at the point of collision and the
     method returns true.
 
     The method returns false if no collision is detected.
@@ -1320,7 +1464,7 @@ CosmoSmashEngine::detectCollisionsWithExplosions()
 }
 
 
-void CosmoSmashEngine::detectCollisionsBetweenFallingObjects()
+void CosmoSmashEngine::detectCollisionsBetweenFallingObjects(SpriteList &slist)
 /*
     Sprite lists containing killed sprites will contain null
     elements after this method finishes.  The caller is
@@ -1328,21 +1472,23 @@ void CosmoSmashEngine::detectCollisionsBetweenFallingObjects()
 */
 {
     SpriteList::iterator it;
-    for (it = pulsarSprites.begin(); it != pulsarSprites.end(); it++)
+    for (it = slist.begin(); it != slist.end(); it++)
     {
-	const Sprite *aPulsar = *it;
-	if (aPulsar == NULL)
+	const Sprite *s = *it;
+	if (s == NULL)
 	    continue;
 	
-	bool coll = detectCollisionsWithBullet(*aPulsar, rockSprites, 0);
+	bool coll = false;
 	if (!coll)
-	    coll = detectCollisionsWithBullet(*aPulsar, saucerSprites, 0);
+	    coll = detectCollisionsWithBullet(*s, rockSprites, 0);
 	if (!coll)
-	    coll = detectCollisionsWithBullet(*aPulsar, saucerBulletSprites, 0);
+	    coll = detectCollisionsWithBullet(*s, saucerSprites, 0);
+	if (!coll)
+	    coll = detectCollisionsWithBullet(*s, saucerBulletSprites, 0);
 	if (coll)
 	{
 	    *it = NULL;
-	    delete aPulsar;
+	    delete s;
 	}
     }
 }
@@ -1356,6 +1502,7 @@ CosmoSmashEngine::detectCollisions()
 */
 {
     SpriteList newRocks;
+    bool playerBoosted = false;
 
     /*  Detect collisions involving the player's bullets:
     */
@@ -1386,6 +1533,8 @@ CosmoSmashEngine::detectCollisions()
 		    */
 		    if (rand() % 3 == 0)
 		    {
+			playShortExplosionSound();
+
 			pair<Sprite *, Sprite *> p = splitBigRock(*rock);
 			newRocks.push_back(p.first);
 			newRocks.push_back(p.second);
@@ -1396,8 +1545,6 @@ CosmoSmashEngine::detectCollisions()
 			    two new small rocks.  @sarrazip 20010530
 			*/
 			explode = false;
-			eraseSprite(*rock);
-			eraseSprite(*bullet);
 		    }
 		}
 		else
@@ -1461,17 +1608,23 @@ CosmoSmashEngine::detectCollisions()
 			    *bullet, saucerBulletSprites, SCORE_SAUCER_BULLET))
 	    *itb = NULL;
 
+	if (detectCollisionsWithBullet(
+			    *bullet, questionSprites, SCORE_QUESTION))
+	{
+	    *itb = NULL;
+	    playerBoosted = true;
+	}
+
 	if (*itb == NULL)  // if bullet hit something
 	    delete bullet;
 
     }   // for each bullet
 
 
-    detectCollisionsBetweenFallingObjects();
+    detectCollisionsBetweenFallingObjects(pulsarSprites);
+    detectCollisionsBetweenFallingObjects(questionSprites);
 
-
-    if (detectCollisionsWithExplosions())
-	return;  // if player dies, no need to continue
+    (void) detectCollisionsWithExplosions();
 
     removeNullElementsFromSpriteList(baseBulletSprites);
     removeNullElementsFromSpriteList(rockSprites);
@@ -1479,8 +1632,25 @@ CosmoSmashEngine::detectCollisions()
     removeNullElementsFromSpriteList(pulsarSprites);
     removeNullElementsFromSpriteList(saucerSprites);
     removeNullElementsFromSpriteList(saucerBulletSprites);
+    removeNullElementsFromSpriteList(questionSprites);
 
     rockSprites.insert(rockSprites.end(), newRocks.begin(), newRocks.end());
+
+    if (playerBoosted)
+	boostPlayer();
+}
+
+
+void
+CosmoSmashEngine::boostPlayer()
+{
+    playerBoostTicks = rnd(100, 150);
+
+    switch (rnd(0, 1))
+    {
+	case 0: playerBoostType = TRIPLE_BULLETS; break;
+	case 1: playerBoostType = CLOSER_BULLETS; break;
+    }
 }
 
 
@@ -1532,9 +1702,12 @@ static bool gamma_table_initialized = false;
 static void
 wu_line(SDL_Surface *surface,
 		       Uint32 x0, Uint32 y0, Uint32 x1, Uint32 y1,
-		       Uint32 fgc, Uint32 bgc, int nlevels, int nbits)
+		       Uint32 fgc, Uint32 bgc)
 /* Source: http://mail.lokigames.com/ml/sdl/0288.html */
 {
+	const int nlevels = 256;
+	const int nbits = 8;
+
   if (!gamma_table_initialized)
   {
     /* generate gamma correction table */
@@ -1669,7 +1842,7 @@ static void
 line(SDL_Surface *surface,
 		Uint32 x1, Uint32 y1, Uint32 x2, Uint32 y2, Uint32 color)
 {
-    wu_line(surface, x1, y1, x2, y2, color, 0, 256, 8);
+    wu_line(surface, x1, y1, x2, y2, color, 0);
 }
 
 
@@ -1679,14 +1852,29 @@ line(SDL_Surface *surface,
 void
 CosmoSmashEngine::restoreBackground()
 {
+    // Background:
+    SDL_Rect rect0 = { 0, 0, theDrawingPixmapSize.x, theDrawingPixmapSize.y };
+    (void) SDL_FillRect(theSDLScreen, &rect0, blackColor);
+
     // Ground:
-    SDL_Rect rect = { 0, groundPos, theDrawingPixmapSize.x, groundHeight };
-    (void) SDL_FillRect(theSDLScreen, &rect, greenColor);
+    SDL_Rect rect1 = { 0, groundPos, theDrawingPixmapSize.x, groundHeight };
+    (void) SDL_FillRect(theSDLScreen, &rect1, greenColor);
 
     // Stars:
-    CoupleList::const_iterator it;
+    bool moveStars = (numLives > 0 && tickCount % 2 == 0);
+    CoupleList::iterator it;
     for (it = starPositions.begin(); it != starPositions.end(); it++)
-	putpixel(theSDLScreen, (*it).x, (*it).y, whiteColor);
+    {
+	Couple &pos = *it;
+	putpixel(theSDLScreen, pos.x, pos.y, whiteColor);
+
+	if (moveStars)
+	{
+	    pos.y--;
+	    if (pos.y < 0)
+		pos.y = mountainTopPos - 1;
+	}
+    }
 
     // Mountain range:
     Couple previous(-1, -1);
@@ -1705,12 +1893,23 @@ void
 CosmoSmashEngine::drawSprites()
 {
     if (playerIsActive())
-	copySpritePixmap(*playerSprite, 0, playerSprite->getPos());
+    {
+	int pixmapIndex = (playerBoostTicks > 0 ? ((tickCount & 4) != 0) : 0);
+	copySpritePixmap(*playerSprite, pixmapIndex, playerSprite->getPos());
+    }
 
     SpriteList::const_iterator it;
     for (it = rockSprites.begin(); it != rockSprites.end(); it++)
 	copySpritePixmap(**it, (**it).currentPixmapIndex, (**it).getPos());
 
+    if (useGameExtensions)
+	for (it = questionSprites.begin(); it != questionSprites.end(); it++)
+	{
+	    size_t &index = (**it).currentPixmapIndex;
+	    index = (tickCount / 8) % 2;
+	    copySpritePixmap(**it, index, (**it).getPos());
+	}
+
     for (it = spinnerSprites.begin(); it != spinnerSprites.end(); it++)
     {
 	size_t &index = (**it).currentPixmapIndex;
@@ -1718,6 +1917,8 @@ CosmoSmashEngine::drawSprites()
 	copySpritePixmap(**it, index, (**it).getPos());
     }
 
+    if (!pulsarSprites.empty() && tickCount % (NUM_PULSAR_IMAGES * 4) == 0)
+	playSoundEffect(pulsarBeepSound);
     for (it = pulsarSprites.begin(); it != pulsarSprites.end(); it++)
     {
 	size_t &index = (**it).currentPixmapIndex;
@@ -1749,36 +1950,25 @@ CosmoSmashEngine::drawSprites()
 	copySpritePixmap(**it, (**it).currentPixmapIndex, (**it).getPos());
     }
 
-    if (updateScore)
     {
+	int ext = (useGameExtensions ? '*' : ' ');
 	char s[64];
-	snprintf(s, sizeof(s), "Score      : %010ld", theScore);
+	snprintf(s, sizeof(s), "Score      :%8ld%c", theScore, ext);
 	writeString(s, scoreAreaPos - Couple(0, 15));
-	snprintf(s, sizeof(s), "Peak Score : %010ld", thePeakScore);
+	snprintf(s, sizeof(s), "Peak Score :%8ld%c", thePeakScore, ext);
 	writeString(s, scoreAreaPos);
 	updateScore = false;
     }
 
-    if (updateNumLives)
     {
 	char s[64];
-	snprintf(s, sizeof(s), "Lives:%4d", numLives);
+	snprintf(s, sizeof(s), "Lives:%3d", numLives);
 	writeString(s, numLivesAreaPos);
 	updateNumLives = false;
     }
 }
 
 
-void
-CosmoSmashEngine::eraseSprite(const Sprite &s)
-{
-    Couple pos = s.getPos();
-    Couple size = s.getSize();
-    SDL_Rect rect = { pos.x, pos.y, size.x, size.y };
-    (void) SDL_FillRect(theSDLScreen, &rect, blackColor);
-}
-
-
 void
 CosmoSmashEngine::addToScore(long n)
 {
@@ -1788,7 +1978,8 @@ CosmoSmashEngine::addToScore(long n)
 	/*  Maintain the peak score. Give the player a new life each time
 	    the peak score crosses some boundary.
 	*/
-	static const long newLifeScoreInterval = 1000;
+	const long newLifeScoreInterval =
+		(useGameExtensions ? 10000 : 1000);
 	long before = thePeakScore / newLifeScoreInterval;
 	thePeakScore = theScore;
 	long after = thePeakScore / newLifeScoreInterval;
@@ -1846,25 +2037,77 @@ CosmoSmashEngine::displayStartMessage(bool display)
 	displayMessage("");
     else
     {
-	static unsigned long firstTick = tickCount;
+	static const Couple fontdim = getFontDimensions();
 
-	if ((tickCount - firstTick) % 64 < 32)
-	    displayMessage("Cosmosmash - by Pierre Sarrazin");
-	else
-	    displayMessage("Press SPACE to start");
+	displayMessage("Cosmosmash " VERSION " - by Pierre Sarrazin",
+			scoreAreaPos.y - fontdim.y);
+	displayMessage("       Distributed under the GNU GPL       ",
+			scoreAreaPos.y);
+
+
+	static const char *lines[] =
+	{
+	    "Press SPACE to start a standard game    ",
+	    "Press E to start a game with extensions ",
+	    "                                        ",
+	    "During game: move with left/right arrows",
+	    "and press down arrow for hyperspace;    ",
+	    "press P to pause the game.              ",
+	    NULL
+	};
+	static int numLines = 6;  // CAUTION: must be right number of lines
+
+	static int y = (groundPos - fontdim.y * numLines) / 2;
+	for (size_t i = 0; lines[i] != NULL; i++)
+	    displayMessage(lines[i], y + i * fontdim.y);
     }
 }
 
 
 void
 CosmoSmashEngine::displayMessage(const char *msg)
+{
+    displayMessage(msg, scoreAreaPos.y);
+}
+
+
+void
+CosmoSmashEngine::displayMessage(const char *msg, int y)
+/*  Displays the string 'msg' at a horizontally centered position
+    and at vertical position 'y'.
+    The font used is that of gengameng.
+*/
 {
     static const Couple fontdim = getFontDimensions();
-    static const char *blank = "                                      ";
 
     int x = (theScreenSizeInPixels.x - strlen(msg) * fontdim.x) / 2;
 
-    if (msg != blank)  // first erase the region, then write the string
-	displayMessage(blank);
-    writeString(msg, Couple(x, scoreAreaPos.y));
+    writeString(msg, Couple(x, y));
+}
+
+
+void
+CosmoSmashEngine::playShortExplosionSound()
+{
+    int soundNo = rand() % NUM_ROCK_HIT_SOUNDS;
+    playSoundEffect(rockHitSounds[soundNo]);
+}
+
+
+void
+CosmoSmashEngine::playSoundEffect(SoundMixer::Chunk &wb)
+{
+    /*  'theSoundMixer' is still NULL if the sound effects were not wanted.
+    */
+    if (theSoundMixer != NULL)
+    {
+	try
+	{
+	    theSoundMixer->playChunk(wb);
+	}
+	catch (const SoundMixer::Error &e)
+	{
+	    fprintf(stderr, "playSoundEffect: %s\n", e.what().c_str());
+	}
+    }
 }
